from dataclasses import dataclass, field, fields
from typing import ClassVar

YL_SENSOR_BIT_GYRO: int = 1
YL_SENSOR_BIT_ACCEL: int = 2
YL_SENSOR_BIT_MAG: int = 4
YL_SENSOR_BIT_BARO: int = 8
YL_SENSOR_BIT_TEMP: int = 16
YL_SENSOR_BIT_GPS: int = 32

@dataclass
class YL_EEPTS_INPUT_DATA:
    gyro_data: list[float] = None  # XYZ in radians/sec. only needs to be float32
    accel_data: list[float] = None # XYZ in G force units(9.8 m/s^2). only needs to be float32
    mag_data: list[float] = None  # XYZ in Gauss. only needs to be float32
    orient_data: list[float] = None # XYZW Optional, may not be here if YL_BUILTIN_ORIENT is false
    barometer_data: float = 0  # in hPa. only needs to be float32
    temperature: float = 0  # in degrees C. only needs to be float32
    gps_longitude: float = 0  # in decimal degrees (positive east, negative west). needs to be float64
    gps_latitude: float = 0  # in decimal degrees (positive north, negative south). needs to be float64
    gps_altitude: float = 0  # in meters. only needs to be float32
    timestamp: int = 0 # in microseconds. uint32_t

#Locomotion Modes
YL_LOCOMOTION_IDLE: int = 0
YL_LOCOMOTION_WALKING: int = 1
YL_LOCOMOTION_JOGGING: int = 2
YL_LOCOMOTION_RUNNING: int = 3
YL_LOCOMOTION_CRAWLING: int = 4
YL_LOCOMOTION_OTHER: int = 6
YL_LOCOMOTION_UNKNOWN: int = 5

#Sensor Locations
YL_SENSOR_RSHOULDER: int = 5
YL_SENSOR_BACK: int = 3
YL_SENSOR_RHAND: int = 4
YL_SENSOR_WAIST: int = 2
YL_SENSOR_CHEST: int = 1
YL_SENSOR_UNKNOWN: int = 0

LocomotionModes = {
    YL_LOCOMOTION_IDLE: "Idle",
    YL_LOCOMOTION_WALKING: "Walking",
    YL_LOCOMOTION_JOGGING: "Jogging",
    YL_LOCOMOTION_RUNNING: "Running",
    YL_LOCOMOTION_CRAWLING: "Crawling",
    YL_LOCOMOTION_OTHER: "Other",
    YL_LOCOMOTION_UNKNOWN: "Unknown"
}

SensorLocations = {
    YL_SENSOR_RSHOULDER: "RShoulder",
    YL_SENSOR_BACK: "Back",
    YL_SENSOR_RHAND: "RHand",
    YL_SENSOR_WAIST: "Waist",
    YL_SENSOR_CHEST: "Chest",
    YL_SENSOR_UNKNOWN: "Unknown"
}

PresetMotionWR = 0
PresetMotionWRC = 1

PresetHandDisabled = 0
PresetHandEnabled = 1
PresetHandOnly = 2

PresetHeadingDynamic = 0
PresetHeadingStatic = 1

@dataclass
class YL_EEPTS_OUTPUT_DATA:

    segment_count: int = 0                  #number of segments (a segment refers to a single step action)
    timestamp: int = 0                      #in microseconds, set to end of last estimated segment
    estimated_gps_longitude: float = 0      #DDMM.MMMM (positive east, negative west)
    estimated_gps_latitude: float = 0       #DDMM.MMMM (positive north, negative south)
    estimated_gps_altitude: float = 0       #in meters
    estimated_heading_angle: float = 0      #in degrees (0 north, 90 east, 180 south, 270 west)
    estimated_distance_travelled: float = 0 #in meters, since last reset (A reset is a start command)
    estimated_distance_x: float = 0         #in meters since last update, along positive east/negative west (An update is a read)
    estimated_distance_y: float = 0         #in meters, since last update, along positive north/negative south
    estimated_distance_z: float = 0         #in meters, since last update, along positive up/negative down
    estimated_locomotion_mode: int = 0      #0=idle, 1=walking, etc...
    estimated_receiver_location: int = 0    #0=unknown, 1=chest, etc...
    last_event_confidence: float = 1
    overall_confidence: float = 1

    LOCMOTION_DICT: ClassVar[dict[int, str]] = { YL_LOCOMOTION_IDLE : "Idle", YL_LOCOMOTION_WALKING : "Walking", YL_LOCOMOTION_JOGGING : "Jogging", YL_LOCOMOTION_RUNNING : "Running", YL_LOCOMOTION_CRAWLING: "Crawling", YL_LOCOMOTION_OTHER: "Other", YL_LOCOMOTION_UNKNOWN: "Unknown" } 
    LOCATION_DICT: ClassVar[dict[int, str]] = { YL_SENSOR_UNKNOWN : "Unknown", YL_SENSOR_CHEST : "Chest", YL_SENSOR_WAIST : "Waist", YL_SENSOR_BACK : "Back", YL_SENSOR_RHAND : "RHand", YL_SENSOR_RSHOULDER : "RShoulder" }

    def clone(self, other: "YL_EEPTS_OUTPUT_DATA"):
        self.segment_count = other.segment_count
        self.timestamp = other.timestamp
        self.estimated_gps_longitude = other.estimated_gps_longitude
        self.estimated_gps_latitude = other.estimated_gps_latitude
        self.estimated_gps_altitude = other.estimated_gps_altitude
        self.estimated_heading_angle = other.estimated_heading_angle
        self.estimated_distance_travelled = other.estimated_distance_travelled
        self.estimated_distance_x = other.estimated_distance_x
        self.estimated_distance_y = other.estimated_distance_y
        self.estimated_distance_z = other.estimated_distance_z
        self.estimated_locomotion_mode = other.estimated_locomotion_mode
        self.estimated_receiver_location = other.estimated_receiver_location
        self.last_event_confidence = other.last_event_confidence
        self.overall_confidence = other.overall_confidence

    def get_locomotion_string(self):
        return YL_EEPTS_OUTPUT_DATA.LOCMOTION_DICT[self.estimated_locomotion_mode]
    
    def get_location_string(self):
        return YL_EEPTS_OUTPUT_DATA.LOCATION_DICT[self.estimated_receiver_location]

    def __str__(self):
        return f"{self.segment_count},{self.timestamp}," \
               f"{self.estimated_gps_longitude},{self.estimated_gps_latitude}," \
               f"{self.estimated_gps_altitude},{self.estimated_heading_angle}," \
               f"{self.estimated_distance_travelled},{self.estimated_distance_x}," \
               f"{self.estimated_distance_y},{self.estimated_distance_z}," \
               f"{self.estimated_locomotion_mode},{self.estimated_receiver_location}"
    
    def print_fancy(self):
        print("Segment Count:", self.segment_count)
        print("Timestamp:", self.timestamp)
        print("Longitude:", self.estimated_gps_longitude)
        print("Latitude:", self.estimated_gps_latitude)
        print("Altitude:", self.estimated_gps_altitude)
        print("Heading:", self.estimated_heading_angle)
        print("Distance Travelled:", self.estimated_distance_travelled)
        print("Delta_X:", self.estimated_distance_x)
        print("Delta_Y:", self.estimated_distance_y)
        print("Delta_Z:", self.estimated_distance_z)
        print("Locomotion:", self.get_locomotion_string())
        print("Sensor Location:", self.get_location_string())
        print("Confidence:", self.last_event_confidence)
        print("Overall Confidence:", self.overall_confidence)

@dataclass
class SensorData:
    accel_x : float
    accel_y : float
    accel_z : float

    gyro_x : float
    gyro_y : float
    gyro_z : float

    mag_x : float
    mag_y : float
    mag_z : float

    accel : list[float] = field(init=False, default_factory=list)
    gyro : list[float] = field(init=False, default_factory=list)
    mag : list[float] = field(init=False, default_factory=list)

    def __post_init__(self):
        self.accel = [self.accel_x, self.accel_y, self.accel_z]
        self.gyro = [self.gyro_x, self.gyro_y, self.gyro_z]
        self.mag = [self.mag_x, self.mag_y, self.mag_z]
    
    def __str__(self):
        return f"{self.accel_x},{self.accel_y},{self.accel_z}," \
               f"{self.gyro_x},{self.gyro_y}, {self.gyro_z}," \
               f"{self.mag_x},{self.mag_y},{self.mag_z}"

    def print_fancy(self):
        print("Accel:", self.accel)
        print("Gyro:", self.gyro)
        print("Mag:", self.mag)

@dataclass
class DebugMessage:
    level: int = 0
    module: int = 0
    msg: str = ""

    def get_display_str(self):
        return f"Level: {self.level} Module: {self.module}    {self.msg}"

@dataclass
class Segment:
    #Output Structure
    segment_count: int = 0
    timestamp: int = 0
    estimated_gps_longitude: float = 0
    estimated_gps_latitude: float = 0
    estimated_gps_altitude: float = 0
    estimated_heading_angle: float = 0
    estimated_distance_travelled: float = 0
    estimated_distance_x: float = 0
    estimated_distance_y: float = 0
    estimated_distance_z: float = 0
    estimated_locomotion_mode: int = 0
    estimated_receiver_location: int = 0
    last_event_confidence: float = 0
    overall_confidence: float = 0

    #Segment Info Structure
    start_global_index: int = 0
    end_global_index: int = 0
    f_start_index_offset: float = 0
    f_end_index_offset: float = 0
    len: int = 0
    f_len: float = 0
    dir: int = 0

    #Debug Messages:
    debug_msgs: list[DebugMessage] = field(default_factory=lambda: [])
    
    @classmethod
    def from_only_output_obj(cls, output: YL_EEPTS_OUTPUT_DATA):
        new_obj = cls()
        output_fields = [f for f in dir(output) if not f.startswith('__') and not callable(getattr(output, f))]
        for field in output_fields: #Can NOT just assign. That will cause unintentional problems if ever saving segment info
            setattr(new_obj, field, getattr(output, field))

        return new_obj
    
    def __str__(self):
        s = ""
        for field in fields(self):
            attr = getattr(self, field.name)
            if field.name == "estimated_locomotion_mode":
                s += f"{field.name}: {LocomotionModes[attr]}\n"
            elif field.name == "estimated_receiver_location":
                s += f"{field.name}: {SensorLocations[attr]}\n"
            else:
                s += f"{field.name}: {attr}\n"
        return s


from yostlabs.tss3.api import ThreespaceSensor
from yostlabs.communication.serial import ThreespaceSerialComClass
from yostlabs.tss3.utils.version import ThreespaceFirmwareUploader

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

uploader = ThreespaceFirmwareUploader(sensor, verbose=True)
uploader.set_firmware_path("embedded_2024_dec_20.xml")

def upload_callback(percent: float):
    print(f"Percent Done: {percent:.2f}%")

#Unnecessary, but nice to use
uploader.set_percent_callback(upload_callback)
uploader.upload_firmware()

#Even after uploading the connection is maintained
print(sensor.getPrimaryCorrectedAccelVec())

sensor.cleanup()


from yostlabs.tss3.api import ThreespaceSensor, StreamableCommands
from yostlabs.communication.serial import ThreespaceSerialComClass
from yostlabs.tss3.utils.streaming import ThreespaceStreamingManager, ThreespaceStreamingStatus
import time

"""
The streaming manager (still has some WIP features) is intended to be used in applications to allow
multiple clients that care about streaming different data to all stream the data they desire at the rate (rate part is WIP) 
they want at the same time. It works by managing registration of streaming items and callbacks.
The DPG suite heavily utilizes this tool.

This is a basic example that only utilizes the basic features.
To see a more complex use of this tool, please look at the source for the open source DPG suite
"""

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

manager = ThreespaceStreamingManager(sensor)

#Everything is registered to an owner object that way something that way each client
#can only register a command and unregister it once
client = object()
manager.register_command(client, StreamableCommands.GetPrimaryCorrectedAccelVec)
manager.register_command(client, StreamableCommands.GetTimestamp)

def stream_callback(status: ThreespaceStreamingStatus):
    if status == ThreespaceStreamingStatus.Data:
        #Can either get the entire last streaming response
        response = manager.get_last_response()

        #Or by type.
        accel = manager.get_value(StreamableCommands.GetPrimaryCorrectedAccelVec)
        timestamp = manager.get_value(StreamableCommands.GetTimestamp)

        #The by type method is to be preferred, as if any other client has registered other commands, this
        #callback would not know how to parse the data part of the response. But requesting by type removes that issue.

        #If the header is desired, get the last response and retrieve its header.
        header = manager.get_last_response().header

        print(f"{timestamp=} {accel=} {header}")
    

manager.register_callback(stream_callback, hz=200) #Stream the data at 200HZ
manager.enable()

#Get results for 5 seconds then stop
start_time = time.time()
while time.time() - start_time < 5:
    manager.update() #Must call the managers update similar to calling sensor.updateStreaming

manager.disable()

sensor.cleanup()



from typing import Generator

class ThreespaceInputStream:

    """
    Reads specified number of bytes. 
    If that many bytes are not available after timeout, less data will be returned
    """
    def read(self, num_bytes: int) -> bytes:
        raise NotImplementedError()
    
    def read_all(self):
        return self.read(self.length)

    def read_until(self, expected: bytes) -> bytes:
        raise NotImplementedError()

    """Allows reading without removing the data from the buffer"""
    def peek(self, num_bytes: int) -> bytes:
        raise NotImplementedError()
    
    def peek_until(self, expected: bytes, max_length: int = None) -> bytes:
        raise NotImplementedError()
    
    def readline(self) -> bytes:
        return self.read_until(b"\r\n")
    
    def peekline(self, max_length: int = None) -> bytes:
        #Lines from the sensor are defined to have a \r\n not just a \n
        return self.peek_until(b"\r\n", max_length=max_length)    
    
    @property
    def length(self) -> int:
        raise NotImplementedError()
    
    @property
    def timeout(self) -> float:
        raise NotImplementedError()

    @timeout.setter
    def timeout(self, timeout: float):
        raise NotImplementedError()    

class ThreespaceOutputStream:

    """Write the given bytes"""
    def write(self, bytes):
        raise NotImplementedError()

class ThreespaceComClass(ThreespaceInputStream, ThreespaceOutputStream):
    """
    Base class for a com class to use with the sensor object.
    Com classes should be initialized without connection and require
    there open called before use
    """
    
    def close(self):
        raise NotImplementedError()
    
    def open(self) -> bool:
        """
        Should return True on success, False on failure
        If already open, should stay open
        """
        raise NotImplementedError()
    
    def check_open(self) -> bool:
        """
        Should return True if the port is currently open, False otherwise.
        Must give the current state, not a cached state
        """
        raise NotImplementedError()
    
    @staticmethod
    def auto_detect() -> Generator["ThreespaceComClass", None, None]:
        """
        Returns a list of com classes of the same type called on nearby
        """
        raise NotImplementedError()
    
    @property
    def reenumerates(self) -> bool:
        """
        If the device Re-Enumerates when going from bootloader to firmware or vice versa, this must return True.
        This indicates to the API that it must search for the new com class representing the object when switching between bootloader and firmware
        """
        raise NotImplementedError
    
    @property
    def name(self) -> str:
        raise NotImplementedError()


from yostlabs.tss3.api import ThreespaceSensor
from yostlabs.communication.ble import ThreespaceBLEComClass

auto_detect = False

if auto_detect:
    #Create a sensor by auto detecting a ThreespaceBLEComClass.
    #It does this by attempting to connect to a device with the Nordic Uart Service
    sensor = ThreespaceSensor(ThreespaceBLEComClass)
else:
    #PUT YOUR SENSORS BLE_NAME HERE
    ble_name = "YL-TSS-####" #Defaults to the lowest 4 hex digits of the sensors serial number
    print("Attempting to discover and connect to a sensor with the name:", ble_name)
    com_class = ThreespaceBLEComClass(ble_name)
    sensor = ThreespaceSensor(com_class)

ble_name = sensor.get_settings("ble_name")
print("Connected to:", ble_name)

result = sensor.getPrimaryCorrectedAccelVec()
print(result)

sensor.cleanup()


from yostlabs.communication.socket import ThreespaceSocketComClass
import time
import threading
import sys
import math

from dataclasses import dataclass
from typing import Callable, Generator

#Bluetooth is an optional install, so we need to handle the case where pybluez2 is not installed
#pip install pybluez2 (May need to do git+https://github.com/airgproducts/pybluez2.git instead for it to work)
#pybluez2 is not included as an optional dependency because the pypi installation is likely to fail.
try:
    import bluetooth
    BLUETOOTH_AVAILABLE = True
except ImportError:
    BLUETOOTH_AVAILABLE = False
    bluetooth = None
    print("pybluez2 is not installed. Bluetooth functionality will be unavailable. Install it with:\n" \
          "    pip install pybluez2\n" \
          "    pip install git+https://github.com/airgproducts/pybluez2.git (May be required for newer versions of python)")

@dataclass
class COD:
    raw: int
    services: list[str]
    major_class: str
    minor_class: str

def decode_class_of_device(cod: int) -> dict:
    """Decode a Bluetooth Class of Device (CoD) integer into its components."""

    # Service Class bit masks (11 bits total)
    services = {
        0x002000: "Limited Discoverable Mode",
        0x004000: "Positioning (Location Identification)",
        0x008000: "Networking",
        0x010000: "Rendering",
        0x020000: "Capturing",
        0x040000: "Object Transfer",
        0x080000: "Audio",
        0x100000: "Telephony",
        0x200000: "Information",
    }

    # Major Device Classes (5 bits)
    major_classes = {
        0x00: "Miscellaneous",
        0x01: "Computer",
        0x02: "Phone",
        0x03: "LAN/Network Access Point",
        0x04: "Audio/Video",
        0x05: "Peripheral",
        0x06: "Imaging",
        0x07: "Wearable",
        0x08: "Toy",
        0x09: "Health",
        0x1F: "Uncategorized",
    }

    # Minor Class mappings for some major classes (for simplicity)
    minor_classes = {
        0x01: {
            0x00: "Uncategorized",
            0x01: "Desktop workstation",
            0x02: "Server-class computer",
            0x03: "Laptop",
            0x04: "Handheld PC/PDA",
            0x05: "Palm-size PC/PDA",
        },
        0x02: {
            0x00: "Uncategorized",
            0x01: "Cellular",
            0x02: "Cordless",
            0x03: "Smartphone",
            0x04: "Wired modem or voice gateway",
            0x05: "Common ISDN access",
        },
        0x05: {
            0x00: "Uncategorized",
            0x01: "Keyboard",
            0x02: "Pointing device",
            0x03: "Combo keyboard/pointing device",
        },
    }

    # Extract fields
    service_bits = cod & 0xFFE000  # top 11 bits
    major_class = (cod >> 8) & 0x1F
    minor_class = (cod >> 2) & 0x3F

    # Decode services
    decoded_services = [name for bit, name in services.items() if service_bits & bit]

    # Decode major class
    major_name = major_classes.get(major_class, "Unknown")

    # Decode minor class (context-dependent)
    minor_name = minor_classes.get(major_class, {}).get(minor_class, f"Minor code {minor_class}")

    return COD(cod, decoded_services or ["None"], major_name, minor_name)

@dataclass
class ScannerResult:
    address: str
    name: str
    class_of_device: COD

class Scanner:

    def __init__(self, desired_scan_time=5):
        self.enabled = False
        self.done = True

        self.continous = False      #If true, will constantly update the nearby devices
        self.execute = False        #Set to true to trigger a read when not in continous mode
        self.nearby = None          #The list of nearby devices
        self.thread = None          #The thread running the asynchronous scanner
        self.updated = False        #Set to true when self.nearby has been updated
        self.duration = math.ceil(desired_scan_time / 1.28) * 1.28  #The time, in second, each scan lasts. Scanning is in 1.28 second intervals, so this may differ from the desired scan time

    def start(self):
        if not self.done:
            self.execute = True
            return
        self.thread = threading.Thread(target=self.process, daemon=True)
        self.enabled = True
        self.done = False
        self.updated = False
        self.thread.start()

    def stop(self):
        self.enabled = False

    def set_continous(self, continous: bool):
        self.continous = continous

    def get_most_recent(self):
        if not self.updated: return None
        self.updated = False
        return self.nearby

    @property
    def is_running(self):
        return self.done

    def process(self):
        if not BLUETOOTH_AVAILABLE:
            raise ImportError("pybluez2 is not installed. Install it with: pip install git+https://github.com/airgproducts/pybluez2.git")
        while self.enabled:
            if not self.continous and not self.execute: continue
            nearby = bluetooth.discover_devices(duration=math.ceil(self.duration/1.28), lookup_names=True, lookup_class=True)
            self.nearby = [ScannerResult(addr, name, decode_class_of_device(cod)) for addr, name, cod in nearby]
            self.execute = False
            self.updated = True
        self.done = True

if sys.platform == "win32":
    import ctypes
    from ctypes import wintypes

    # Define BLUETOOTH_ADDRESS structure
    class BLUETOOTH_ADDRESS(ctypes.Structure):
        _fields_ = [("rgBytes", ctypes.c_byte * 6)]

    # Load Bluetooth API
    win_bluetooth = ctypes.WinDLL("BluetoothAPIs.dll")

    # Define function prototype
    BluetoothRemoveDevice = win_bluetooth.BluetoothRemoveDevice
    BluetoothRemoveDevice.argtypes = [ctypes.POINTER(BLUETOOTH_ADDRESS)]
    BluetoothRemoveDevice.restype = wintypes.BOOL

    def remove_device(mac_address: str):
        # Convert MAC string "XX:XX:XX:XX:XX:XX" -> bytes reversed
        addr_bytes = bytes.fromhex(mac_address.replace(":", ""))[::-1]
        addr = BLUETOOTH_ADDRESS()
        ctypes.memmove(addr.rgBytes, addr_bytes, 6)

        err = BluetoothRemoveDevice(ctypes.byref(addr))
        print("Remove device status:", err)
        if not err:
            print(f"Successfully removed device {mac_address}")
        else:
            if err == 1168:  # ERROR_NOT_FOUND
                print(f"Device {mac_address} not found or already unpaired.")
            else:
                print(f"Failed to remove device {mac_address}. Error code: {err}")
else:
    def remove_device(mac_address: str):
        raise NotImplementedError("Only implemented on windows")

class ThreespaceBluetoothComClass(ThreespaceSocketComClass):

    SCANNER: Scanner = None

    def __init__(self, addr: str, name: str = None, connection_timeout=None):
        if not BLUETOOTH_AVAILABLE:
            raise ImportError("pybluez2 is not installed. Install it with: pip install pybluez2 (May need to do git+https://github.com/airgproducts/pybluez2.git instead for it to work)")
        super().__init__(bluetooth.BluetoothSocket(bluetooth.Protocols.RFCOMM), (addr, 1), connection_timeout=connection_timeout)
        self.address = addr
        self.__name = name or addr

    @property
    def name(self) -> str:
        return self.__name

    @classmethod
    def __lazy_init_scanner(cls):
        if cls.SCANNER is None:
            cls.SCANNER = Scanner()
            cls.SCANNER.start()

    @classmethod
    def set_scanner_continous(cls, continous: bool):
        cls.__lazy_init_scanner()
        cls.SCANNER.set_continous(continous)

    @staticmethod
    def __default_filter(result: ScannerResult):
        return result.class_of_device.major_class == "Wearable" and result.class_of_device.minor_class == "Minor code 0"

    @staticmethod
    def auto_detect(wait_for_update=True, filter: Callable[[ScannerResult],bool] = None) -> Generator["ThreespaceBluetoothComClass", None, None]:
        """
        Returns a list of com classes of the same type called on nearby
        """
        cls = ThreespaceBluetoothComClass
        cls.__lazy_init_scanner()
        if filter is None: 
            filter = cls.__default_filter
        cls.SCANNER.start()
        if wait_for_update:
            cls.SCANNER.updated = False
            while not cls.SCANNER.updated: time.sleep(0.1)
        if cls.SCANNER.nearby is None: return
        for device_info in cls.SCANNER.nearby:
            if not filter(device_info): continue
            name = device_info.name or None
            yield ThreespaceBluetoothComClass(device_info.address, name)

    @staticmethod
    def unpair(address: str):
        """
        It is recommended to call this after done with a ThreespaceBluetoothComClass object
        so that windows will not report the device as still being available when powered off.
        This occurs because windows Bluetooth detect functions report nearby devices & paired devices,
        regardless of their power status. By removing the device from the list of paired devices, auto detect
        will only report actual nearby and powered devices.
        EX: ThreespaceBluetoothComClass.unpair(com.address)
        """
        remove_device(address)


if __name__ == "__main__":
    from yostlabs.tss3.api import ThreespaceSensor

    com = None
    for device in ThreespaceBluetoothComClass.auto_detect():
        com = device

    if com is None:
        print("Failed to detect a bluetooth sensor")
        exit()
    print("Connecting to:", com.name)

    sensor = ThreespaceSensor(com, verbose=True)

    sensor.set_settings(stream_slots=0)
    print(sensor.get_settings("stream_slots", "stream_hz"))
    print(sensor.getTaredOrientation())

    sensor.startStreaming()

    start_time = time.perf_counter()
    while time.perf_counter() - start_time < 5:
        sensor.updateStreaming()
        packet = sensor.getOldestStreamingPacket()
        while packet is not None:
            print(packet)
            packet = sensor.getOldestStreamingPacket()
    sensor.stopStreaming()
    sensor.cleanup()


THREESPACE_HEADER_STATUS_BIT_POS = 0
THREESPACE_HEADER_TIMESTAMP_BIT_POS = 1
THREESPACE_HEADER_ECHO_BIT_POS = 2
THREESPACE_HEADER_CHECKSUM_BIT_POS = 3
THREESPACE_HEADER_SERIAL_BIT_POS = 4
THREESPACE_HEADER_LENGTH_BIT_POS = 5
THREESPACE_HEADER_NUM_BITS = 6

THREESPACE_HEADER_STATUS_BIT = (1 << THREESPACE_HEADER_STATUS_BIT_POS)
THREESPACE_HEADER_TIMESTAMP_BIT = (1 << THREESPACE_HEADER_TIMESTAMP_BIT_POS)
THREESPACE_HEADER_ECHO_BIT = (1 << THREESPACE_HEADER_ECHO_BIT_POS)
THREESPACE_HEADER_CHECKSUM_BIT = (1 << THREESPACE_HEADER_CHECKSUM_BIT_POS)
THREESPACE_HEADER_SERIAL_BIT = (1 << THREESPACE_HEADER_SERIAL_BIT_POS)
THREESPACE_HEADER_LENGTH_BIT = (1 << THREESPACE_HEADER_LENGTH_BIT_POS)

FIRMWARE_VALID_BIT = 0x01

PASSIVE_CALIBRATE_GYRO = (1 << 0)
PASSIVE_CALIBRATE_MAG_REF = (1 << 1)

STREAMING_MAX_HZ = 2000

THREESPACE_OUTPUT_MODE_ASCII = 1
THREESPACE_OUTPUT_MODE_BINARY = 2

THREESPACE_GET_SETTINGS_ERROR_RESPONSE = "<KEY_ERROR>"
THREESPACE_BINARY_SETTINGS_ID_SIZE = 4
THREESPACE_BINARY_READ_SETTINGS_ID = 0xC695B5E1

#This is not comprehensive, just enough to seperate keys from debug messages
THREESPACE_SETTING_KEY_INVALID_CHARS = " :;"

THREESPACE_FILE_STREAMING_MAX_PACKET_SIZE = 512
THREESPACE_LIVE_LOG_STREAM_MAX_PACKET_SIZE = 2048

THREESPACE_SN_FAMILY_POS = 14 * 4
THREESPACE_SN_VARIATION_POS = 11 * 4
THREESPACE_SN_VERSION_POS = 10 * 4
THREESPACE_SN_MAJOR_REVISION_POS = 8 * 4
THREESPACE_SN_MINOR_REVISION_POS = 6 * 4
THREESPACE_SN_INCREMENTOR_POS = 0 * 4

THREESPACE_SN_FAMILY_MSK = 0xFF << THREESPACE_SN_FAMILY_POS
THREESPACE_SN_VARIATION_MSK = 0xFFF << THREESPACE_SN_VARIATION_POS
THREESPACE_SN_VERSION_MSK = 0xF << THREESPACE_SN_VERSION_POS
THREESPACE_SN_MAJOR_REVISION_MSK = 0xFF << THREESPACE_SN_MAJOR_REVISION_POS
THREESPACE_SN_MINOR_REVISION_MSK = 0xFF << THREESPACE_SN_MINOR_REVISION_POS
THREESPACE_SN_INCREMENTOR_MSK = 0xFFFFFF << THREESPACE_SN_INCREMENTOR_POS

THREESPACE_FAMILY_DEV = "DEV"
THREESPACE_FAMILY_USB = "USB"
THREESPACE_FAMILY_WIRELESS = "WL"
THREESPACE_FAMILY_EMBEDDED = "EM"
THREESPACE_FAMILY_BLUETOOTH = "BT"
THREESPACE_FAMILY_DATA_LOGGER = "DL"
THREESPACE_FAMILY_MICRO_USB = "MUSB"
THREESPACE_FAMILY_LX = "LX"

THREESPACE_SN_FAMILY_TO_NAME = {
    0x00 : THREESPACE_FAMILY_DEV,
    0x11 : THREESPACE_FAMILY_USB,
    0x12 : THREESPACE_FAMILY_WIRELESS,
    0x14 : THREESPACE_FAMILY_EMBEDDED,
    0x15 : THREESPACE_FAMILY_BLUETOOTH,
    0x16 : THREESPACE_FAMILY_DATA_LOGGER,
    0x17 : THREESPACE_FAMILY_MICRO_USB,
    0x21: THREESPACE_FAMILY_LX
}

THREESPACE_ERR_COMMIT_FS_LOCKED=8


from yostlabs.tss3.api import ThreespaceSensor, StreamableCommands
from yostlabs.communication.serial import ThreespaceSerialComClass
import time

#Create a sensor by auto detecting a ThreespaceSerialComClass
sensor = ThreespaceSensor(ThreespaceSerialComClass)

#Setup streaming
sensor.set_settings(stream_hz=1000) #caps at 2000
sensor.set_settings(stream_delay=0, stream_duration=0) #stream instantly on command, stream forever

#To set stream slots, provide a comma seperated list of command numbers. Up to 16 commands can be streamed at once.
#Some commands may require a parameter passed to them. To do so, put a ':' after the command number and then the parameter
#The below set_settings call has an example of this.
#255 is considered an empty stream slot
stream_slots_string = f"{StreamableCommands.GetAllPrimaryCorrectedData.value},{StreamableCommands.GetTaredOrientation.value},{StreamableCommands.GetRawAccelVec.value}:{sensor.valid_accels[0]}"
sensor.set_settings(stream_slots=stream_slots_string)
print("Stream Slots:", sensor.get_settings("stream_slots"))

sensor.set_settings(header_timestamp=1) #Enable the timestamp in the header to see the time at which the message was generated
sensor.set_settings(timestamp=0) #Set the timestamp back to 0 so the messages start from 0

#Start the streaming
sensor.startStreaming()

#Get results for 5 seconds then stop
start_time = time.time()
while time.time() - start_time < 5:
    #This must be called repeatedly while streaming to allow the sensor to read in the results
    sensor.updateStreaming()

    #Check for any streaming packets that got read in and print them out.
    packet = sensor.getOldestStreamingPacket()
    while packet is not None:
        print(packet)
        packet = sensor.getOldestStreamingPacket()

sensor.stopStreaming()

sensor.cleanup()



import asyncio
import async_timeout
import threading
import time
from bleak import BleakScanner, BleakClient
from bleak.backends.device import BLEDevice
from bleak.backends.scanner import AdvertisementData
from bleak.backends.characteristic import BleakGATTCharacteristic
from bleak.exc import BleakDeviceNotFoundError
from dataclasses import dataclass

#Services
NORDIC_UART_SERVICE_UUID = "6e400001-b5a3-f393-e0a9-e50e24dcca9e"

#Characteristics
NORDIC_UART_RX_UUID = "6e400002-b5a3-f393-e0a9-e50e24dcca9e"
NORDIC_UART_TX_UUID = "6e400003-b5a3-f393-e0a9-e50e24dcca9e"

DEVICE_NAME_UUID = "00002a00-0000-1000-8000-00805f9b34fb"
APPEARANCE_UUID = "00002a01-0000-1000-8000-00805f9b34fb"

FIRMWARE_REVISION_STRING_UUID = "00002a26-0000-1000-8000-00805f9b34fb"
HARDWARE_REVISION_STRING_UUID = "00002a27-0000-1000-8000-00805f9b34fb"
SERIAL_NUMBER_STRING_UUID = "00002a25-0000-1000-8000-00805f9b34fb"
MANUFACTURER_NAME_STRING_UUID = "00002a29-0000-1000-8000-00805f9b34fb"

@dataclass
class ThreespaceBLENordicUartProfile:
    SERVICE_UUID: str
    RX_UUID: str
    TX_UUID: str

class TssBLENoConnectionError(Exception): ...

def ylBleEventLoopThread(loop: asyncio.AbstractEventLoop):
    asyncio.set_event_loop(loop)
    loop.run_forever()

from yostlabs.communication.base import *
class ThreespaceBLEComClass(ThreespaceComClass):

    DEFAULT_TIMEOUT = 2
    EVENT_LOOP = None
    EVENT_LOOP_THREAD = None

    DEFAULT_PROFILE = ThreespaceBLENordicUartProfile(NORDIC_UART_SERVICE_UUID, NORDIC_UART_RX_UUID, NORDIC_UART_TX_UUID)
    REGISTERED_PROFILES: list[ThreespaceBLENordicUartProfile] = [DEFAULT_PROFILE]

    @classmethod
    def __lazy_event_loop_init(cls):
        if cls.EVENT_LOOP is None:
            cls.EVENT_LOOP = asyncio.new_event_loop()
            cls.EVENT_LOOP_THREAD = threading.Thread(target=ylBleEventLoopThread, args=(cls.EVENT_LOOP,), daemon=True)
            cls.EVENT_LOOP_THREAD.start()

    def __init__(self, ble: BleakClient | BLEDevice | str, discover_name: bool = True, discovery_timeout=10, error_on_disconnect=True, adv: AdvertisementData = None, profile: ThreespaceBLENordicUartProfile=None):
        """
        Parameters
        ----------
        ble : Can be either a BleakClient, BleakDevice, MacAddress String, or localName string
        discover_name : If true, a string ble parameter is interpreted as a localName, else as a MacAddress
        discovery_timeout : Max amount of time in seconds to discover the BLE device for the corresponding MacAddress/localName
        error_on_disconnect : If trying to read while the sensor is disconnected, an exception will be generated. This may be undesirable \
        if it is expected that the sensor will frequently go in and out of range and the user wishes to preserve data (such as streaming)
        """
        self.__lazy_event_loop_init()
        self.adv = adv
        bleak_options = { "timeout": discovery_timeout, "disconnected_callback": self.__on_disconnect }
        if isinstance(ble, BleakClient):    #Actual client
            self.client = ble
            self.__name = ble.address
        elif isinstance(ble, str): 
            if discover_name: #Local Name stirng
                self.__lazy_init_scanner()
                future = asyncio.run_coroutine_threadsafe(BleakScanner.find_device_by_name(ble, timeout=discovery_timeout), self.EVENT_LOOP)
                device = future.result()
                if device is None:
                    raise BleakDeviceNotFoundError(ble)
                self.client = BleakClient(device, **bleak_options)
                self.__name = ble
            else: #Address string
                self.client = BleakClient(ble, **bleak_options)
                self.__name = self.client.address
        elif isinstance(ble, BLEDevice):
            self.client = BleakClient(ble, **bleak_options)
            self.__name = ble.name #Use the local name instead of the address
        else:
            raise TypeError("Invalid type for creating a ThreespaceBLEComClass:", type(ble), ble)

        #Select the profile
        self.profile = profile
        if self.profile is None:
            if self.adv is not None and len(self.adv.service_uuids) > 0:
                for service_uuid in self.adv.service_uuids:
                    self.profile = self.get_profile(service_uuid)
                    if self.profile is not None:
                        break
                if self.profile is None:
                    self.profile = ThreespaceBLEComClass.DEFAULT_PROFILE
                    raise Exception(f"Unknown Service UUIDS: {self.adv.service_uuids}")
            else:
                self.profile = ThreespaceBLEComClass.DEFAULT_PROFILE

        self.__timeout = self.DEFAULT_TIMEOUT

        self.buffer = bytearray()
        self.data_read_event: asyncio.Event = None

        #Default to 20, will update on open
        self.max_packet_size = 20
        
        self.error_on_disconnect = error_on_disconnect

        #is_connected is different from open. Open is the cached idea of whether the BLE connection should be active,
        #while is_connected is the actual state of the connection. Basically, opened means a connection is desired, while
        #is_connected means it is actually connected.
        self.__opened = False

        #client.is_connected is really slow (noticeable when called in bulk, which happens do to the assert_connected)... 
        #So instead using the disconnected callback and this variable to manage tracking the state without the delay
        self.__connected = False

        #Writing functions will naturally throw an exception if disconnected. Reading ones don't because they use notifications rather
        #then direct reads. This means reading functions will need to assert the connection status but writing does not.

    async def __async_open(self):
        self.data_read_event = asyncio.Event()
        await self.client.connect()
        await self.client.start_notify(self.profile.TX_UUID, self.__on_data_received)

    def open(self):
        #If trying to open while already open, this infinitely loops
        if self.__opened: 
            if not self.__connected and self.error_on_disconnect:
                self.close()
            return
        asyncio.run_coroutine_threadsafe(self.__async_open(), self.EVENT_LOOP).result()
        self.max_packet_size = self.client.mtu_size - 3 #-3 to account for the opcode and attribute handle stored in the data packet
        self.__opened = True
        self.__connected = True

    async def __async_close(self):
        #There appears to be a bug where if you call close too soon after is_connected returns false,
        #the disconnect call will hang on Windows. It seems similar to this issue: https://github.com/hbldh/bleak/issues/1359
        await asyncio.sleep(0.5)
        await self.client.disconnect()
        self.data_read_event = None
        self.buffer.clear()

    def close(self):
        if not self.__opened: return
        asyncio.run_coroutine_threadsafe(self.__async_close(), self.EVENT_LOOP).result()
        self.__opened = False

    def __on_disconnect(self, client: BleakClient):
        self.__connected = False

    #Goal is that this is always called after something that would have already performed an async callback
    #to prevent needing to run the event loop. Running the event loop frequently is slow. Which is also why this
    #comclass will eventually have a threaded asyncio version.
    def __assert_connected(self):
        if not self.__connected and self.error_on_disconnect:
            raise TssBLENoConnectionError(f"{self.name} is not connected")

    def check_open(self):
        if not self.__connected and self.__opened and self.error_on_disconnect:
            self.close()
        return self.__connected

    def __on_data_received(self, sender: BleakGATTCharacteristic, data: bytearray):
        self.buffer += data
        self.data_read_event.set()

    def write(self, bytes: bytes):
        start_index = 0
        while start_index < len(bytes):
            end_index = min(len(bytes), start_index + self.max_packet_size) #Can only send max_packet_size data per call to write_gatt_char
            asyncio.run_coroutine_threadsafe(
                self.client.write_gatt_char(self.profile.RX_UUID, bytes[start_index:end_index], response=False),
                self.EVENT_LOOP).result()
            start_index = end_index
    
    async def __await_read(self, timeout_time: int):
        self.__assert_connected()
        try:
            async with async_timeout.timeout_at(timeout_time):
                await self.data_read_event.wait()
            self.data_read_event.clear()
            return True
        except:
            return False

    async def __await_num_bytes(self, num_bytes: int):
        start_time = self.EVENT_LOOP.time()
        while len(self.buffer) < num_bytes and self.EVENT_LOOP.time() - start_time < self.timeout:
            await self.__await_read(start_time + self.timeout)

    def read(self, num_bytes: int):
        asyncio.run_coroutine_threadsafe(self.__await_num_bytes(num_bytes), self.EVENT_LOOP).result()
        num_bytes = min(num_bytes, len(self.buffer))
        data = self.buffer[:num_bytes]
        del self.buffer[:num_bytes]
        return data

    def peek(self, num_bytes: int):
        asyncio.run_coroutine_threadsafe(self.__await_num_bytes(num_bytes), self.EVENT_LOOP).result()
        num_bytes = min(num_bytes, len(self.buffer))
        data = self.buffer[:num_bytes]
        return data        
    
    #Reads until the pattern is received, max_length is exceeded, or timeout occurs
    async def __await_pattern(self, pattern: bytes, max_length: int = None):
        if max_length is None: max_length = float('inf')
        start_time = self.EVENT_LOOP.time()
        while pattern not in self.buffer and self.EVENT_LOOP.time() - start_time < self.timeout and len(self.buffer) < max_length:
            await self.__await_read(start_time + self.timeout)
        return pattern in self.buffer

    def read_until(self, expected: bytes) -> bytes:
        asyncio.run_coroutine_threadsafe(self.__await_pattern(expected), self.EVENT_LOOP).result()
        if expected in self.buffer: #Found the pattern
            length = self.buffer.index(expected) + len(expected)
            result = self.buffer[:length]
            del self.buffer[:length]
            return result
        #Failed to find the pattern, just return whatever is there
        result = self.buffer.copy()
        self.buffer.clear()
        return result

    def peek_until(self, expected: bytes, max_length: int = None) -> bytes:
        asyncio.run_coroutine_threadsafe(self.__await_pattern(expected, max_length=max_length), self.EVENT_LOOP).result()
        if expected in self.buffer:
            length = self.buffer.index(expected) + len(expected)
        else:
            length = len(self.buffer)

        if max_length is not None and length > max_length:
            length = max_length

        return self.buffer[:length]

    @property
    def length(self):
        return len(self.buffer) 

    @property
    def timeout(self) -> float:
        return self.__timeout
    
    @timeout.setter
    def timeout(self, timeout: float):
        self.__timeout = timeout    

    @property
    def reenumerates(self) -> bool:
        return False
    
    @property
    def name(self) -> str | None:
        """
        The name of the device. This may be the Address or the Local Name of the device
        depending on how discovery was done.
        May also be None
        """
        return self.__name
    
    @property
    def address(self) -> str:
        return self.client.address    

    @classmethod
    def get_profile(cls, service_uuid: str):
        for profile in cls.REGISTERED_PROFILES:
            if profile.SERVICE_UUID == service_uuid:
                return profile
        return None

    SCANNER: BleakScanner = None
    SCANNER_LOCK = None
    SCANNER_RUNNING = False

    SCANNER_CONTINOUS = False   #Controls if scanning will continously run
    SCANNER_TIMEOUT = 5         #Controls the scanners timeout
    SCANNER_FIND_COUNT = 1      #When continous=False, will stop scanning after at least this many devices are found. Set to None to search the entire timeout.
    SCANNER_EXPIRATION_TIME = 5 #Controls the timeout for detected BLE sensors. If a sensor hasn't been detected again in this amount of time, its removed from discovered devices

    #Format: Address - dict = { device: ..., adv: ..., last_found: ... }
    discovered_devices: dict[str,dict] = {}

    @classmethod
    def set_profiles(cls, profiles: list[ThreespaceBLENordicUartProfile]):
        cls.REGISTERED_PROFILES = profiles
        if cls.SCANNER is not None:
            asyncio.run_coroutine_threadsafe(cls.create_scanner(), cls.EVENT_LOOP).result()
            cls.__remove_unused_profiles()

    @classmethod
    def register_profile(cls, profile: ThreespaceBLENordicUartProfile):        
        if any(v.SERVICE_UUID == profile.SERVICE_UUID for v in cls.REGISTERED_PROFILES): return
        cls.REGISTERED_PROFILES.append(profile)
        if cls.SCANNER is not None:
            asyncio.run_coroutine_threadsafe(cls.create_scanner(), cls.EVENT_LOOP).result()
    
    @classmethod
    def unregister_profile(cls, service_uuid: str|ThreespaceBLENordicUartProfile):
        if isinstance(service_uuid, ThreespaceBLENordicUartProfile):
            service_uuid = service_uuid.SERVICE_UUID
        index = None
        for i in range(len(cls.REGISTERED_PROFILES)):
            if cls.REGISTERED_PROFILES[i].SERVICE_UUID == service_uuid:
                index = i
                break
        del cls.REGISTERED_PROFILES[index]
        if cls.SCANNER is not None:
            asyncio.run_coroutine_threadsafe(cls.create_scanner(), cls.EVENT_LOOP).result()
            cls.__remove_unused_profiles()

    @classmethod
    def __remove_unused_profiles(cls):
        if cls.SCANNER is None: return
        to_remove = []
        valid_service_uuids = [v.SERVICE_UUID for v in cls.REGISTERED_PROFILES]
        with cls.SCANNER_LOCK:
            for address in cls.discovered_devices:
                adv: AdvertisementData = cls.discovered_devices[address]["adv"]
                if not any(uuid in valid_service_uuids for uuid in adv.service_uuids):
                    to_remove.append(address)
            for address in to_remove:
                del cls.discovered_devices[address]

    #Scanner should be created inside of the Async Context that will use it
    @classmethod
    async def create_scanner(cls):
        uuids = [v.SERVICE_UUID for v in cls.REGISTERED_PROFILES]
        restart_scanner = cls.SCANNER_RUNNING
        with cls.SCANNER_LOCK:
            if restart_scanner:
                await cls.__async_stop_scanner()
            cls.SCANNER = BleakScanner(detection_callback=cls.__detection_callback, service_uuids=uuids)
            if restart_scanner:
                await cls.__async_start_scanner()

    @classmethod
    def __lazy_init_scanner(cls):
        cls.__lazy_event_loop_init()
        if cls.SCANNER is None:
            cls.SCANNER_LOCK = threading.Lock()
            asyncio.run_coroutine_threadsafe(cls.create_scanner(), cls.EVENT_LOOP).result()

    @classmethod
    def __detection_callback(cls, device: BLEDevice, adv: AdvertisementData):
        with cls.SCANNER_LOCK:
            cls.discovered_devices[device.address] = {"device": device, "adv": adv, "last_found": time.time()}
    
    @classmethod
    async def __async_start_scanner(cls):
        if cls.SCANNER_RUNNING: return
        await cls.SCANNER.start()
        cls.SCANNER_RUNNING = True

    @classmethod
    async def __async_stop_scanner(cls):
        if not cls.SCANNER_RUNNING: return
        await cls.SCANNER.stop()
        cls.SCANNER_RUNNING = False        

    @classmethod
    def __start_scanner(cls):
        if cls.SCANNER_RUNNING: return
        asyncio.run_coroutine_threadsafe(cls.__async_start_scanner(), cls.EVENT_LOOP).result()
    
    @classmethod
    def __stop_scanner(cls):
        if not cls.SCANNER_RUNNING: return
        asyncio.run_coroutine_threadsafe(cls.__async_stop_scanner(), cls.EVENT_LOOP).result()
        cls.__stop_scanner()

    @classmethod
    def set_scanner_continous(cls, continous: bool):
        """
        If not using continous mode, functions like update_nearby_devices and auto_detect are blocking with the following rules:
        - Will search for at most SCANNER_TIMEOUT time
        - Will stop searching immediately once SCANNER_FIND_COUNT is reached

        If using continous mode, no scanning functions are blocking. However, the user must continously call 
        update_nearby_devices to ensure up to date information.
        """
        cls.__lazy_init_scanner()
        cls.SCANNER_CONTINOUS = continous
        if continous: 
            cls.__start_scanner()
        else: 
            cls.__stop_scanner()

    @classmethod
    def update_nearby_devices(cls):
        """
        Updates ThreespaceBLEComClass.discovered_devices using the current configuration.
        """
        cls.__lazy_init_scanner()
        if cls.SCANNER_CONTINOUS:
            with cls.SCANNER_LOCK:
                #Remove expired devices
                cur_time = time.time()
                to_remove = [] #Avoiding concurrent list modification
                for device in cls.discovered_devices:
                    if cur_time - cls.discovered_devices[device]["last_found"] > cls.SCANNER_EXPIRATION_TIME:
                        to_remove.append(device) 
                for device in to_remove:
                    del cls.discovered_devices[device]
                discovered = cls.discovered_devices.copy()
        else:
            #Mark all devices as invalid before searching for nearby devices
            cls.discovered_devices.clear()
            start_time = time.time()
            end_time = cls.SCANNER_TIMEOUT or float('inf')
            end_count = cls.SCANNER_FIND_COUNT or float('inf')
            asyncio.run_coroutine_threadsafe(cls.SCANNER.start(), cls.EVENT_LOOP).result()
            while time.time() - start_time < end_time and len(cls.discovered_devices) < end_count:
                time.sleep(0)
            asyncio.run_coroutine_threadsafe(cls.SCANNER.stop(), cls.EVENT_LOOP).result()
            discovered = cls.discovered_devices.copy()
        return discovered
    
    @classmethod
    def get_discovered_nearby_devices(cls):
        """
        A helper to get a copy of the discovered devices
        """
        with cls.SCANNER_LOCK:
            discovered = cls.discovered_devices.copy()
        return discovered

    @staticmethod
    def auto_detect() -> Generator["ThreespaceBLEComClass", None, None]:
        """
        Returns a list of com classes of the same type called on nearby.
        These ports will start unopened. This allows the caller to get a list of ports without having to connect.
        """
        cls = ThreespaceBLEComClass
        cls.update_nearby_devices()
        # Copy devices while holding lock, then yield without holding lock to avoid deadlock
        # if the user decides to interact with the com class inside the loop
        with cls.SCANNER_LOCK:
            devices_snapshot = list(cls.discovered_devices.values())
        for device_info in devices_snapshot:
            yield(ThreespaceBLEComClass(device_info["device"], adv=device_info["adv"]))
